import logging
from typing import Any

from grimoiressg.arguments import parse_arguments_to_initial_context
from grimoiressg.config import read_config
from grimoiressg.content_files import recursively_read_files
from grimoiressg.context import Context
from grimoiressg.modules import available_modules
from grimoiressg.utils import logger


def apply_modules(data: Any, config: dict[str, Any], context: Context):
    for module in config.get("enabled_modules", []):
        logger.info("Applying module %s...", module)

        available_modules[module](data, context, config)


def main():
    context = parse_arguments_to_initial_context()
    config = read_config(context)

    data = recursively_read_files(context)
    apply_modules(data, config, context)

    logger.info("Done.")
    logging.shutdown()



from collections.abc import Callable
import glob
import os
from typing import Any


def to_relative(path: str):
    trimmed = path.removeprefix(os.getcwd())
    if trimmed != path:
        trimmed = "." + trimmed
    return trimmed


def for_each_glob(glob_path: str, callback: Callable[[str], list[dict[str, Any]]]):
    results = []

    for filename in glob.glob(os.path.realpath(glob_path)):
        results.extend(callback(filename))

    return results



import os
import pathlib
from typing import Any

import yaml
from yaml import Loader

from grimoiressg.context import Context
from grimoiressg.utils import logger, for_each_glob, to_relative

possible_separators = (
    "---\n",
    "+++\n",
)


def handle_file(filename: str) -> list[dict[str, Any]]:
    logger.debug(" Reading %s...", to_relative(filename))

    file_suffix = pathlib.Path(filename).suffix
    match file_suffix:
        case ".yml" | ".yaml":
            data = handle_yaml(filename)
        case ".md":
            data = handle_markdown(filename)
        case _:
            raise ValueError(f"File extension {file_suffix} not supported.")

    data["filename"] = filename
    data["relative_filename"] = to_relative(filename)

    results = [data]

    relative_dir = os.path.dirname(filename)
    for filename in data.get("include", []):
        filename = relative_dir + "/" + filename

        sub_data = for_each_glob(filename, handle_file)

        results.extend(sub_data)

    return results


def handle_markdown(filename: str) -> dict[str, Any]:
    parsed_data: dict[str, Any] = {}

    with open(filename, "r") as file:
        meta_data = ""
        content = ""

        if content := file.read():
            meta_data, content = split_frontmatter(content)

        if meta_data:
            parsed_data = yaml.safe_load(meta_data)

        if content != "":
            parsed_data["markdown"] = content

    return parsed_data


def handle_yaml(filename: str) -> dict[str, Any]:
    with open(filename, "r") as file:
        return yaml.load(file, Loader)


def expect_separator(input: str) -> str | None:
    separator_candidate = input[0:4]
    if separator_candidate in possible_separators:
        return separator_candidate

    return None


def split_frontmatter(input: str) -> tuple[str | None, str]:
    separator = expect_separator(input)
    if not separator:
        return None, input

    frontmatter_end = input.find(separator, 4)
    if frontmatter_end == -1:
        # frontmatter never ends -> there is no front matter
        return None, input

    return input[4:frontmatter_end], input[frontmatter_end + 4 :]


def deduplicate(candidates: list[Any]):
    names = set()
    results = []

    for candidate in candidates:
        if candidate["relative_filename"] not in names:
            names.add(candidate["relative_filename"])
            results.append(candidate)

    return results


def recursively_read_files(context: Context):
    data: list[dict[str, Any]] = []

    logger.info("Reading content files...")

    for filename in context.filenames:
        data.extend(for_each_glob(filename, handle_file))

    data = deduplicate(data)

    logger.info(f"Read {len(data)} files in total.")

    return data



import gzip
import os
from itertools import batched
from xml.etree import ElementTree as ET

from grimoiressg.context import Context
from grimoiressg.utils import to_relative, logger

INDEX_FILE_STRATEGY_NONE = "none"
INDEX_FILE_STRATEGY_AUTO = "auto"


def sitemaps_default_config():
    return {
        "file_prefix": "sitemap",
        "loc_prefix": "https://example.com/",
        "index_file_strategy": INDEX_FILE_STRATEGY_AUTO,
        "compression": False,
    }


def get_files_to_map(data, sitemap_config):
    content_for_sitemap = filter(
        lambda item: item.get("output", False) and not item.get("skip_sitemap", False),
        data,
    )

    if sitemap_config["index_file_strategy"] == INDEX_FILE_STRATEGY_AUTO:
        # maximum number of entries is 50 000, however there is also a 50 MiB size limit
        # -> make 20 000 item batches - to be safe
        return list(batched(content_for_sitemap, 20000))
    else:
        return [content_for_sitemap]


def get_sitemap_file_suffix(sitemap_config):
    if sitemap_config["compression"]:
        return ".xml.gz"
    else:
        return ".xml"


def save_sitemaps_file(xml_data, name, context: Context, sitemap_config):
    xml_str = ET.tostring(xml_data, encoding="utf8")

    filename = os.path.realpath(
        context.output_dir + "/" + name + get_sitemap_file_suffix(sitemap_config)
    )
    logger.debug("Writing sitemap %s", to_relative(filename))

    open_function = gzip.open if sitemap_config["compression"] else open
    with open_function(filename, "wb") as file:
        file.write(xml_str)


def generate_index_file(context, sitemap_config, number_of_batches):
    root = ET.Element(
        "sitemapindex",
        attrib={
            "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            "xsi:schemaLocation": "http://www.sitemaps.org/schemas/sitemap/0.9 "
            "http://www.sitemaps.org/schemas/sitemap/0.9/siteindex.xsd",
            "xmlns": "http://www.sitemaps.org/schemas/sitemap/0.9",
        },
    )

    for i in range(1, number_of_batches + 1):
        sitemap = ET.SubElement(root, "sitemap")
        loc = ET.SubElement(sitemap, "loc")
        loc.text = (
            sitemap_config["loc_prefix"]
            + sitemap_config["file_prefix"]
            + str(i)
            + get_sitemap_file_suffix(sitemap_config)
        )

    save_sitemaps_file(root, sitemap_config["file_prefix"], context, sitemap_config)


def generate_sitemaps_file(batch, name, context, sitemap_config):
    root = ET.Element(
        "urlset",
        attrib={
            "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            "xsi:schemaLocation": "http://www.sitemaps.org/schemas/sitemap/0.9 "
            "http://www.sitemaps.org/schemas/sitemap/0.9/sitemap.xsd",
            "xmlns": "http://www.sitemaps.org/schemas/sitemap/0.9",
        },
    )

    for entry in batch:
        url = ET.SubElement(root, "url")
        loc = ET.SubElement(url, "loc")
        loc.text = sitemap_config["loc_prefix"] + entry["output"]

    save_sitemaps_file(root, name, context, sitemap_config)


def generate_sitemaps(data, context, config):
    sitemaps_config = sitemaps_default_config()
    sitemaps_config.update(config.get("sitemaps", {}))

    batches = get_files_to_map(data, sitemaps_config)
    if len(batches) > 1:
        logger.info("Entry limit exceeded; generating index file...")
        generate_index_file(context, sitemaps_config, len(batches))
        for i, batch in enumerate(batches):
            generate_sitemaps_file(
                batch,
                sitemaps_config["file_prefix"] + str(i + 1),
                context,
                sitemaps_config,
            )
    if len(batches) == 1:
        generate_sitemaps_file(
            batches[0], sitemaps_config["file_prefix"], context, sitemaps_config
        )



from typing import Any
from grimoiressg.context import Context
from grimoiressg.utils import logger


def extract_tags(data: list[dict[str, Any]], context: Context, config: dict[str, Any]):
    tags = {}

    for entry in data:
        for tag in entry.get("tags", []):
            entry_list = tags.get(tag, [])
            entry_list.append(entry)
            tags[tag] = entry_list

    if tags:
        logger.debug("Found tags:")
        for tag in tags.keys():
            logger.debug(" - %s (%d files)", tag, len(tags[tag]))
    else:
        logger.debug("No tags found.")

    context.tags = tags




from grimoiressg.modules.markdown import compile_markdown
from grimoiressg.modules.sitemaps import generate_sitemaps
from grimoiressg.modules.tags import extract_tags
from grimoiressg.modules.templating import render_templates

available_modules = {
    "tags": extract_tags,
    "markdown": compile_markdown,
    "templating": render_templates,
    "sitemaps": generate_sitemaps,
}


def load_external_module(module):
    __import__(module)



import markdown

from grimoiressg.utils import logger


def compile_markdown(data, context, config):
    for entry in data:
        if "markdown" in entry:
            logger.debug("Compiling markdown for %s...", entry['relative_filename'])
            entry["markdown_compiled"] = markdown.markdown(entry["markdown"], extensions=['tables'])



import os

from grimoiressg.context import Context
from jinja2 import Environment, FileSystemLoader

from grimoiressg.utils import to_relative, logger

from dataclasses import asdict

jinja_env = Environment(loader=FileSystemLoader("/"))


def render_templates(data, context: Context, _config):
    files_written = 0
    # render templates in reverse order, so included renderings can be used

    for entry in reversed(data):
        if "template" in entry:
            template_path = os.path.realpath(
                os.path.dirname(entry["filename"]) + "/" + entry["template"]
            )

            template_dir = os.path.dirname(template_path)
            logger.debug("Rendering template for %s...", entry["relative_filename"])

            template = jinja_env.get_template(template_path)
            entry["rendered"] = template.render(
                **asdict(context), current=entry, all=data, template_dir=template_dir
            )

        if "rendered" in entry and "output" in entry:
            files_written += 1
            filename = os.path.realpath(context.output_dir + "/" + entry["output"])

            logger.debug(" writing to %s", to_relative(filename))
            os.makedirs(os.path.dirname(filename), exist_ok=True)

            with open(filename, "w") as file:
                file.write(entry["rendered"])

    logger.debug("%d rendered", files_written)



import argparse

from grimoiressg.context import Context
from grimoiressg.utils import logger


def parse_arguments_to_initial_context() -> Context:
    parser = argparse.ArgumentParser(
        description="""
            Grimoire is a minimalistic Static Site Generator.
            In the simplest case the only argument needed is at least one content file. \
            The rest of the flags is used to customize the behavior.
        """
    )
    _ = parser.add_argument("content_file", nargs="+", help="one or more content files")
    _ = parser.add_argument(
        "-o",
        "--output",
        default="./output/",
        help="the output directory (default: ./output/)",
    )
    _ = parser.add_argument(
        "-c", "--config", help="the config file to use", default=None
    )

    args, _ = parser.parse_known_args()

    context = Context(
        output_dir=args.output,
        config_file=args.config,
        filenames=args.content_file,
    )

    logger.debug("Output directory: %s", context.output_dir)
    logger.debug("Config file: %s", context.config_file)
    logger.debug("Content files:")

    for filename in context.filenames:
        logger.debug(" - %s", filename)

    return context



import logging

import yaml
from yaml import Loader

from grimoiressg.context import Context
from grimoiressg.modules import available_modules, load_external_module
from grimoiressg.utils import logger


def default_config() -> dict[str, list[str]]:
    return {
        "enabled_modules": [
            "tags",
            "markdown",
            "templating",
        ],
    }


def read_config(context: Context):
    config_file = context.config_file

    if not config_file:
        logger.info("No config file given; using default config")

        config = default_config()

    else:
        logger.info("Loading config file...")

        with open(config_file, "r") as file:
            config = yaml.load(file, Loader) or {}

    for module in config.get("load_modules", []):
        logger.debug(" Loading external module %s", module)

        load_external_module(module)

    logger.debug("Enabled modules:")

    for module in config.get("enabled_modules", []):
        logger.debug(" - %s", module)

        if module not in available_modules:
            logger.critical("Module does not exist: %s", module)
            logging.shutdown()
            exit(1)

    return config



from dataclasses import dataclass, field
from typing import Any


@dataclass
class Context:
    output_dir: str
    config_file: str | None
    filenames: str
    enabled_modules: list[str] = field(default_factory=lambda: [])
    tags: dict[str, Any] = field(default_factory=lambda: {})
